#!/usr/bin/env python

# Copyright (c) 2015 Aleksey Maksimov and Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os
from os.path import join as jp
import json
import argparse

import requests
from bottle import route, run, static_file, post, response

_HERE = os.path.abspath(os.path.dirname(__file__))

_json_dir = '/var/www/jenkinsflow/'
jenkins_url = 'http://localhost:8080/'


@route('/jenkinsflow/graph')
def index():
    return static_file('flow_vis.html', root=_HERE)


@route('/jenkinsflow/js/<filename>')
def js(filename):
    return static_file(filename, root=jp(_HERE, 'js'))


@route('/jenkinsflow/stylesheets/<filename>')
def stylesheets(filename):
    return static_file(filename, root=jp(_HERE, 'stylesheets'))


@post('/jenkinsflow/builds')
def builds():
    rsp = requests.get(jenkins_url + 'queue/api/json')
    response.content_type = 'application/json'
    return json.dumps(rsp.json())


@post('/jenkinsflow/build/<name>')
def job(name):
    rsp = requests.get(jenkins_url + 'job/' + name + '/api/json')
    response.content_type = 'application/json'
    return json.dumps(rsp.json())


@route('/jenkinsflow/flow_graph.json')
def graph_json():
    response.content_type = 'text'
    return open(jp(_json_dir, 'flow_graph.json'), 'r', encoding="utf-8")
    # return static_file('flow_graph.json', root=_json_dir)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Serve flow graph.')
    parser.add_argument('--hostname', default='localhost', help='hostname to listen on')
    parser.add_argument('--port', default=9090, help='port to listen on')
    parser.add_argument('--json-dir', default=_json_dir, help='Where to find json flow graph')
    args = parser.parse_args()
    _json_dir = args.json_dir

    run(host=args.hostname, port=args.port, debug=True)



# Copyright (c) 2012 - 2024 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys
import os
import signal
import multiprocessing

import setproctitle


class LoggingProcess(multiprocessing.Process):
    proc_name_prefix = "jenkinsflow_script_api_"

    def __init__(self, group=None, target=None, output_file_name=None, workspace=None, name=None, args=(), env=None):
        self.user_target = target
        super().__init__(group=group, target=self.run_job_wrapper, name=name, args=args)
        self.output_file_name = output_file_name
        self.workspace = workspace
        self.env = env
        self._build_res_unstable = False

    def run_job_wrapper(self, *args):
        setproctitle.setproctitle(self.proc_name_prefix + self.name)

        # Set signalhandler for changing job result
        def set_result(_sig, _frame):
            print("\nGot SIGUSR1: Changing result to 'unstable'")
            self._build_res_unstable = True
        signal.signal(signal.SIGUSR1, set_result)

        os.chdir(self.workspace)
        os.environ.update(self.env)
        os.environ['EXECUTOR_NUMBER'] = repr(self.pid)

        try:
            rc = self.user_target(*args)
        except Exception as ex:  # pylint: disable=broad-except
            print("jenkinsflow.script_api: Caught exception from job script:", ex)
            rc = 1

        if self._build_res_unstable:
            sys.exit(2)
        sys.exit(rc)

    def run(self):
        sys.stdout = sys.stderr = open(self.output_file_name, 'w', buffering=1, encoding="utf-8")
        super().run()



import os

from jenkinsflow.jenkins_api import JenkinsApi

from . import demo_security as security


def get_jenkins_api():
    url = os.environ.get('JENKINS_URL') or "http://localhost:8080"
    jenkins_api = JenkinsApi(url, username=security.username, password=security.password) if security.default_use_login else JenkinsApi(url)
    return jenkins_api, security.securitytoken



# Copyright (c) 2019 - 2024 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from .ordered_enum import OrderedEnum


class Checking(OrderedEnum):
    MUST_CHECK = 0
    HAS_UNCHECKED = 1
    FINISHED = 2



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys
import time
import subprocess
from pathlib import Path

from . import api_select
from .logger import log, logt
from .cfg import ApiType, AllCfg, opt_strs_to_test_cfg, test_cfg_to_opt_strs


def _abort(log_file, test_file_name, api_type, fixed_prefix, job_name, sleep_time, test_cfg: AllCfg):
    log(log_file, '\n')
    logt(log_file, "Subprocess", __file__)
    logt(log_file, f"Waiting {sleep_time} seconds to abort job:", job_name)
    logt(log_file, "args:", test_file_name, api_type, fixed_prefix, job_name, sleep_time, test_cfg)
    time.sleep(sleep_time)
    logt(log_file, "sleep", sleep_time, "finished")
    with api_select.api(test_file_name, api_type, fixed_prefix='jenkinsflow_test__' + fixed_prefix + '__', login=True, options=test_cfg) as api:
        api.job(job_name, max_fails=0, expect_invocations=0, expect_order=None)
    logt(log_file, "job defined", api)
    api.poll()
    logt(log_file, "polled")
    api.quick_poll()
    logt(log_file, "quick polled")

    abort_me = api.get_job(api.job_name_prefix + job_name)
    logt(log_file, "Abort job:", abort_me)
    abort_me.stop_all()
    logt(log_file, "Aborted")


if __name__ == '__main__':
    job_name = sys.argv[4]
    with open(job_name + '.log', 'a+', encoding="utf-8") as log_file:
        try:
            _abort(
                log_file, sys.argv[1], ApiType[sys.argv[2]], sys.argv[3], job_name, int(sys.argv[5]),
                test_cfg = opt_strs_to_test_cfg(
                    direct_url=sys.argv[6], load_jobs=sys.argv[7], delete_jobs=sys.argv[8], mock_speedup=sys.argv[9], apis_str=sys.argv[10]))
        except Exception as ex:
            print(ex, file=log_file)
            raise


def abort(api, job_name, sleep_time, test_cfg: AllCfg):
    """Call this script as a subprocess"""
    if api.api_type == ApiType.MOCK:
        return

    args = [sys.executable, "-m", f"jenkinsflow.test.framework.{Path(__file__).stem}",
            api.file_name, api.api_type.name, api.func_name.replace('test_', ''), job_name, str(sleep_time),
            *test_cfg_to_opt_strs(test_cfg, api.api_type)]
    with open(job_name + '.log', 'w', encoding="utf-8") as log_file:
        logt(log_file, "Current dir:", Path.cwd())
        logt(log_file, "Invoking abort subprocess.", args)
    subprocess.Popen(args, start_new_session=True)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import pytest

from jenkinsflow.flow import serial
from jenkinsflow.jenkins_api import JenkinsApi
from jenkinsflow.api_base import InvalidJobNameException


from .framework.cfg import ApiType


@pytest.mark.apis(ApiType.JENKINS)
@pytest.mark.parametrize("org_name", ["codeberg-org-jenkinsflow-test", "github-org-jenkinsflow-test"])
def test_invalid_job_in_org_folder_name(api_type, org_name):
    url = "http://localhost:8080"
    api = JenkinsApi(url)
    with pytest.raises(InvalidJobNameException):
        with serial(api, timeout=20, report_interval=1) as ctrl1:
            # Name must not contain '/job/'
            ctrl1.invoke(f"{org_name}/job/jenkinsflow-gh-folder-test/job/main")



"""Utilities used by both nox and pytest invocation"""

import argparse

from .cfg import ApiType, Urls


OPT_API = "--api"
OPT_JOB_LOAD = "--job-load"
OPT_JOB_DELETE = "--job-delete"
OPT_DIRECT_URL = "--direct-url"
OPT_MOCK_SPEEDUP = "--mock-speedup"


def add_options(parser):
    def _add_opt(opt, **kwargs):
        """Pytest uses it's own parser which works similar to ArgParse"""
        if isinstance(parser, argparse.ArgumentParser):
            parser.add_argument(opt, **kwargs)
            return

        # metavar="NAME",
        parser.addoption(opt, **kwargs)

    _add_opt(
        OPT_API,
        action="store",
        default=','.join(at.name for at in list(ApiType)),
        help=f"Comma separated list of APIs to test. Default is all defined apis: {','.join([at.name for at in list(ApiType)])}",
    )

    _add_opt(
        OPT_JOB_LOAD,
        action="store_true",
        default=True,
        help="Load Jenkins jobs - can be skipped to speed up testing if there are no changes to jobs and they are already loaded.",
    )

    _add_opt(
        OPT_JOB_DELETE,
        action="store_true",
        default=False,
        help="Delete Jenkins jobs before loading. May be necessary depending the changes to jobs (or for cleanup of obsolete jobs).",
    )

    _add_opt(
        OPT_DIRECT_URL,
        action="store",
        default=Urls.default_direct_url,
        help="Direct URL (i.e. non-proxied) of Jenkins server.",
    )

    _add_opt(
        OPT_MOCK_SPEEDUP,
        action="store",
        type=int,
        default=500,
        help="Time speedup for mock API tests. If set too high test may fail. If set low it just takes longer to run tests.",
    )



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from ...framework import api_select


def create_jobs(api_type):
    api = api_select.api(__file__, api_type)
    api.flow_job()
    api.job('passwd_args', exec_time=0.5, max_fails=0, expect_invocations=1, expect_order=1,
            params=(('s1', 'no-secret', 'desc'), ('passwd', 'p2', 'desc'), ('PASS', 'p3', 'desc')))
    return api


if __name__ == '__main__':
    create_jobs(api_select.ApiType.JENKINS)



#!/usr/bin/env python3

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

# Demonstrate that the flow can be dynamically calculated by the script

import sys, os, tempfile
import logging

from jenkinsflow.flow import serial
from jenkinsflow.unbuffered import UnBuffered

import lib.get_jenkins_api


# Buffered output does not work well in Jenkins, so in case
# this is run from a Jenkins job, we want unbuffered output.
sys.stdout = UnBuffered(sys.stdout)


def main(api, securitytoken):
    logging.basicConfig()
    logging.getLogger("").setLevel(logging.WARNING)

    print("Doing stuff before flow ...")
    demo_name = 'jenkinsflow_demo__calculated_flow'

    graph_output_dir = os.path.join(tempfile.gettempdir(), demo_name + '_jenkins')
    if not os.path.exists(graph_output_dir):
        os.makedirs(graph_output_dir)

    g1_components = range(1)
    g2_components = range(2)
    g3_components = range(2)
    component_groups = {
        "g1": g1_components,
        "g2": g2_components,
        "g3": g3_components,
    }

    # Flow
    with serial(api, timeout=70, securitytoken=securitytoken, job_name_prefix=demo_name + '__', report_interval=3,
                # Write json flow graph to display in browser, see INSTALL.md
                json_dir=graph_output_dir, json_indent=4) as ctrl1:

        ctrl1.invoke('prepare')

        with ctrl1.parallel(timeout=0, report_interval=3) as ctrl2:
            for gname, group in component_groups.items():
                with ctrl2.serial(timeout=0, report_interval=3) as ctrl3:
                    for component in group:
                        ctrl3.invoke('deploy_component_' + gname + '_' + str(component))

        with ctrl1.parallel(timeout=0, report_interval=3) as ctrl2:
            ctrl2.invoke('report_deploy')
            ctrl2.invoke('prepare_tests')

        with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
            with ctrl2.serial(timeout=40, report_interval=3) as ctrl3:
                ctrl3.invoke('test_ui')
                with ctrl3.parallel(timeout=0, report_interval=3) as ctrl4:
                    for gname, group in component_groups.items():
                        for component in group:
                            ctrl4.invoke('test_component_' + gname + '_' + str(component))
            ctrl2.invoke('test_x')

        ctrl1.invoke('report', password='Y', s1='tst_regression', c1='complete')
        ctrl1.invoke('promote')

    print("Doing stuff after flow ...")


if __name__ == '__main__':
    main(*lib.get_jenkins_api.get_jenkins_api())



from enum import Enum


class KillType(Enum):
    NONE = 0
    CURRENT = 1
    ALL = 2

    def __bool__(self):
        return self != KillType.NONE

    # Python2 compatibility
    __nonzero__ = __bool__



#!/usr/bin/env python3

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

# Just a simple flow

from jenkinsflow.flow import serial

import lib.get_jenkins_api


def main(api, securitytoken):
    demo_name = 'jenkinsflow_demo__basic'

    with serial(api, timeout=70, securitytoken=securitytoken, job_name_prefix=demo_name + '__', report_interval=3) as outer_ctrl:
        outer_ctrl.invoke('prepare')
        outer_ctrl.invoke('deploy_component')

        with outer_ctrl.parallel(timeout=0, report_interval=3) as report_prepare_ctrl:
            report_prepare_ctrl.invoke('report_deploy')
            report_prepare_ctrl.invoke('prepare_tests')

        with outer_ctrl.parallel(timeout=0, report_interval=3) as test_ctrl:
            test_ctrl.invoke('test_x')
            test_ctrl.invoke('test_y')

        outer_ctrl.invoke('report', password='Y', s1='tst', c1='complete')


if __name__ == '__main__':
    main(*lib.get_jenkins_api.get_jenkins_api())



# Copyright (c) 2012 - 2017 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from .utils import base_url_and_api


def set_build_description(
        description: str, replace: bool = False, separator: str = '\n',
        username: str = None, password: str = None,
        build_url: str = None, job_name: str = None, build_number: int = None,
        direct_url: str = None):
    """Utility method to set/append build description on a job build.

    If this is used from inside the Jenkins job you do not have to specify 'build_url' or 'job_name', 'build_number'. You do not have to specify 'direct_url' either
    but may still want to do so if JENKINS_URL points to a proxy, so that rest calls can go directly to Jenkins.
    The 'build_url' is preferred over 'job_name' and 'build_number'.

    Args:
        description: The description to set on the build.
        replace: Replace existing description, if any, instead of appending.
        separator: A separator to insert between any existing description and the new 'description' if 'replace' is not specified.
        username: User Name for Jenkin authentication with secured Jenkins.
        password: Password of Jenkins User.
        build_url: The URL of the jenkins build - preferably non-proxied. Default is os.environ['BUILD_URL'].
        job_name: Name of the job to modify a build on. Default is os.environ['JOB_NAME'].
        build_number: Build Number to modify. . Default is os.environ['BUILD_NUMBER'].
        direct_url: Jenkins URL - preferably non-proxied. If not specified, the value of JENKINS_URL environment variables will be used.
    """

    base_url, api = base_url_and_api(direct_url)
    jenkins = api(direct_uri=base_url, username=username, password=password)
    jenkins.set_build_description(description, replace, separator, build_url, job_name, build_number)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys
import importlib
# import importlib.machinery
from pathlib import Path

import pytest
from pytest import raises

from jenkinsflow.flow import parallel, JobControlFailException

_HERE = Path(__file__).absolute().parent
_DEMO_JOBS_DIR = _HERE/"demos/jobs"

from demo import basic, calculated_flow, prefix, hide_password, errors

from .framework import api_select
from .framework.cfg import ApiType


def load_demo_jobs(demo, api_type):
    print("\nLoad jobs for demo:", demo.__name__)
    simple_demo_name = demo.__name__.replace("demo.", "")
    job_load_module_name = ".demos.jobs." + simple_demo_name + '_jobs'
    job_load = importlib.import_module(job_load_module_name, "test")
    api = job_load.create_jobs(api_type)
    flow_job_name = simple_demo_name + "__0flow"
    return flow_job_name


def _test_demo(demo, api_type):
    print("_test_demo", demo, api_type)
    flow_job_name = load_demo_jobs(demo, api_type)

    with api_select.api(__file__, api_type, fixed_prefix="jenkinsflow_demo__") as api:
        api.job(flow_job_name, max_fails=0, expect_invocations=1, expect_order=1)

        with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
            ctrl1.invoke(flow_job_name)


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_basic(api_type):
    _test_demo(basic, api_type)


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_calculated_flow(api_type):
    _test_demo(calculated_flow, api_type)


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_prefix(api_type):
    _test_demo(prefix, api_type)


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_hide_password(api_type):
    _test_demo(hide_password, api_type)


@pytest.mark.not_apis(ApiType.SCRIPT)  # TODO: script api is not configured to run demos
def test_demos_with_errors(api_type):
    flow_job_name = load_demo_jobs(errors, api_type)

    with api_select.api(__file__, api_type, fixed_prefix="jenkinsflow_demo__") as api:
        api.job(flow_job_name, max_fails=1, expect_invocations=1, expect_order=1)

        with raises(JobControlFailException):
            with parallel(api, timeout=70, job_name_prefix=api.job_name_prefix) as ctrl1:
                ctrl1.invoke(flow_job_name)



#!/usr/bin/env python3

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial

import lib.get_jenkins_api


def main(api, securitytoken):
    with serial(api, timeout=70, securitytoken=securitytoken, job_name_prefix='jenkinsflow_demo__hide_password__', report_interval=3, secret_params='.*PASS.*|.*pass.*') as ctrl:
        # NOTE: In order to ensure that passwords are not displayed in a stacktrace you must never put a literal password
        # In the last line in the with statement, or in any statement that may raise an exception. You shold not really
        # put clear text paswords in you code anyway :)
        p1, p2, p3 = 'SECRET', 'sec', 'not_security'
        ctrl.invoke('passwd_args', password=p1, s1='no-secret', passwd=p2, PASS=p3)


if __name__ == '__main__':
    main(*lib.get_jenkins_api.get_jenkins_api())



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os
import os.path
import time
import urllib.parse
# import json

from .api_base import BuildResult, Progress, ClientError, UnknownJobException, BaseApiMixin, ApiInvocationMixin
from .speed import Speed
from .rest_api_wrapper import ResourceNotFound, RequestsRestApi


_SUPERSEDED_PSEUDO_BUILD_NUM = -1
_DEQUEUED_PSEUDO_BUILD_NUM = -2

_CT_URL_ENC = {'Content-Type': 'application/x-www-form-urlencoded'}

# Query info and parameter definitions of a specific job
_GIVEN_JOB_QUERY_WITH_PARAM_DEFS = "lastBuild[number,result,building],queueItem[why],actions[parameterDefinitions[name,type]],property[parameterDefinitions[name,type]]"

# Initial query
# Build a three level query to handle github organization folder jobs.
# Note that 'name' in response does not contain first 'job/' from url, but does contain intermediate 'job/', so:
# 'name': 'github-org-jenkinsflow-test/job/jenkinsflow-gh-folder-test/job/main'
# Corresponds:
# 'url': 'http://<host>:<port>/job/github-org-jenkinsflow-test/job/jenkinsflow-gh-folder-test/job/main'
_ONE_LEVEL_JOBS_INFO_QUERY = f"jobs[name,{_GIVEN_JOB_QUERY_WITH_PARAM_DEFS}_RECURSE_]"
_TWO_LEVELS_JOBS_INFO_QUERY = _ONE_LEVEL_JOBS_INFO_QUERY.replace("_RECURSE_", "," + _ONE_LEVEL_JOBS_INFO_QUERY)
_THREE_LEVELS_JOBS_INFO_QUERY = _TWO_LEVELS_JOBS_INFO_QUERY.replace("_RECURSE_", "," + _ONE_LEVEL_JOBS_INFO_QUERY)
_FULL_QUERY = f"{_THREE_LEVELS_JOBS_INFO_QUERY.replace('_RECURSE_', '')},primaryView[url]"


# Query status only of a specific job
_GIVEN_JOB_QUICK_QUERY = "lastBuild[number,result,building],queueItem[why]"

# Quick query
# Build a three level query to handle github organization folder job status.
_ONE_LEVEL_JOBS_STATUS_QUERY = f"jobs[name,{_GIVEN_JOB_QUICK_QUERY}_RECURSE_]"
_TWO_LEVELS_JOBS_STATUS_QUERY = _ONE_LEVEL_JOBS_STATUS_QUERY.replace("_RECURSE_", "," + _ONE_LEVEL_JOBS_STATUS_QUERY)
_THREE_LEVELS_JOBS_STATUS_QUERY = _TWO_LEVELS_JOBS_STATUS_QUERY.replace("_RECURSE_", "," + _ONE_LEVEL_JOBS_STATUS_QUERY)
_QUICK_QUERY = _THREE_LEVELS_JOBS_STATUS_QUERY.replace('_RECURSE_', '')


def _result_and_progress(build_dct):
    building = build_dct.get("building")
    result = build_dct["result"]
    progress = Progress.RUNNING if (building or result is None) else Progress.IDLE
    result = BuildResult.UNKNOWN if result is None else BuildResult[result]
    return (result, progress)


class JenkinsApi(Speed, BaseApiMixin):
    """Optimized minimal set of methods needed for jenkinsflow to access Jenkins jobs.

    Args:
        direct_uri (str): Should be a non-proxied uri if possible (e.g. http://localhost:<port> if flow job is running on 'built-in' node)
            The public URI will be retrieved from Jenkins and used in output.
        job_prefix_filter (str): Jobs with names that don't start with this string, will be skpped when polling Jenkins.
        username (str): Name of user authorized to execute all jobs in flow.
        password (str): Password of user.
        invocation_class (class): Defaults to `Invocation`. You can subclass that to provide your own class.
        csrf (bool): Will attempt to get (and use) a CSRF protection crumb from Jenkins. A 404 - ResourceNotFound error is silently
            ignored as this indicates that csrf protection is not enabled on Jenkins.
    """

    def __init__(self, direct_uri, job_prefix_filter=None, username=None, password=None, invocation_class=None, rest_access_provider=RequestsRestApi, csrf=True):
        if username or password:
            if not (username and password):
                raise ValueError("You must specify both username and password or neither")
        self.rest_api = rest_access_provider(direct_uri, username, password)

        self.direct_uri = direct_uri
        self.username = username
        self.password = password
        self.invocation_class = invocation_class or Invocation
        self.job_prefix_filter = job_prefix_filter
        self.jenkins_prefix = urllib.parse.urlsplit(direct_uri).path  # When jenkins is started with a 'prefix' value
        self._public_uri = None
        self.jobs = None
        self.queue_items = {}
        self.is_jenkins = False
        self.csrf = csrf
        self._crumb = None

    def _get_fresh_crumb(self):
        """Get the CSRF crumb to be put on subsequent requests"""
        try:
            crumb = self.rest_api.get_content('/crumbIssuer/api/xml', xpath='concat(//crumbRequestField,":",//crumb)').split(b':')
            self._crumb = {crumb[0]: crumb[1]}
        except ResourceNotFound:
            self.csrf = False

    def get_json(self, url="", **params):  # pylint: disable=inconsistent-return-statements
        # Sometimes (but rarely) Jenkins will Abort the connection when jobs are being aborted!
        for ii in (1, 2, 3):
            try:
                return self.rest_api.get_json(url, **params)
            except ConnectionError as ex:
                if ii == 3:
                    raise
                print("WARNING: Retrying failed 'poll':", ex)
                time.sleep(0.1)

    def post(self, url, payload=None, headers=None, **params):  # pylint: disable=inconsistent-return-statements
        for crumb_attempt in (0, 1):
            if self._crumb:
                if headers:
                    headers = headers.copy()
                    headers.update(self._crumb)
                else:
                    headers = self._crumb

            try:
                return self.rest_api.post(url, payload, headers, **params)
            except ClientError as ex:
                if crumb_attempt:
                    raise
                if self.csrf:
                    if self._crumb:
                        print("INFO: getting new crumb:", ex)
                    self._get_fresh_crumb()

    def headers(self):
        return self.rest_api.headers()

    @property
    def public_uri(self):
        if not self._public_uri:
            query = "primaryView[url]"
            dct = self.get_json(tree=query)
            self._public_uri = dct['primaryView']['url'].rstrip('/')
        return self._public_uri

    def _public_job_url(self, job_name):
        return self.public_uri + '/job/' + job_name

    def _resolve_job_levels(self, dct, parent_job_name, quick: bool):
        """Resolve jobs for a poll. E.g. gh folder jobs have three levels <org>/<repo>/<branch>."""
        has_children = False
        for job_dct in dct.get('jobs') or []:
            has_children = True

            job_name = str(job_dct['name'])
            if self.job_prefix_filter and not job_name.startswith(self.job_prefix_filter):
                continue

            if parent_job_name:
                job_name = f"{parent_job_name}/job/{job_name}"
            has_grand_children = self._resolve_job_levels(job_dct, job_name, quick)

            if quick:
                job = self.jobs.get(job_name)
                if job:
                    job.dct = job_dct
                    continue

                # A new job was created while flow was running, get the remaining properties
                try:
                    job_dct = self.get_json("/job/" + job_name, tree=_GIVEN_JOB_QUERY_WITH_PARAM_DEFS)
                    # print("new job:", job_dct)
                    job = ApiJob(self, job_dct, job_name, has_children=has_grand_children)
                    self.jobs[job_name] = job
                except ResourceNotFound:  # pragma: no cover
                    # Ignore this, the job came and went
                    continue

                continue

            self.jobs[job_name] = ApiJob(self, job_dct, job_name, has_grand_children)
            #if job_dct["_class"] not in ("hudson.model.FreeStyleProject", "org.jenkinsci.plugins.workflow.job.WorkflowJob"):
            #    print(json.dumps(self.get_json(depth=3), indent=2))

        return has_children

    def poll(self):
        if not self.is_jenkins:
            # Determine whether we are talking to Jenkins
            head_response = self.headers()
            if head_response.get("X-Jenkins"):
                self.is_jenkins = True
            else:
                raise Exception("Not connected to Jenkins. Expected X-Jenkins header, got: " + repr(head_response))

        dct = self.get_json(tree=_FULL_QUERY)
        self._public_uri = dct['primaryView']['url'].rstrip('/')

        self.jobs = {}
        self._resolve_job_levels(dct, None, quick=False)

    def quick_poll(self):
        dct = self.get_json(tree=_QUICK_QUERY)
        # print("dct:", json.dumps(dct, indent=2))
        self._resolve_job_levels(dct, None, quick=True)

    def queue_poll(self):
        query = "items[task[name],id]"
        dct = self.get_json("/queue", tree=query)

        queue_items = {}
        for qi_dct in dct.get('items') or []:
            job_name = str(qi_dct['task']['name'])
            if self.job_prefix_filter and not job_name.startswith(self.job_prefix_filter):
                continue

            # print("queue_item name:", job_name)
            queue_items.setdefault(job_name, []).append(qi_dct['id'])
        self.queue_items = queue_items

    def get_job(self, name):
        try:
            return self.jobs[name]
        except KeyError as ex:
            # print("self.jobs:", self.jobs.keys())
            raise UnknownJobException(self._public_job_url(name)) from ex

    def create_job(self, job_name, config_xml):
        self.post('/createItem', name=job_name,
                  headers={'Content-Type': 'application/xml header;charset=utf-8'},
                  payload=config_xml.encode('utf-8'))

    def delete_job(self, job_name):
        try:
            self.post('/job/' + job_name + '/doDelete')
        except ResourceNotFound as ex:
            # TODO: Check error
            raise UnknownJobException(self._public_job_url(job_name), ex) from ex

    def _set_description(self, description, build_url, replace=False, separator: str = '\n'):
        if not replace:
            dct = self.get_json(build_url, tree="description")
            existing_description = dct['description']
            if existing_description:
                description = existing_description + separator + description

        self.post(build_url + '/submitDescription', headers=_CT_URL_ENC, payload={'description': description})

    def set_build_description(
            self, description: str, replace: bool = False, separator: str = '\n',
            build_url: str = None, job_name: str = None, build_number: int = None):
        """Utility to set/append build description.

        Args:
            description:  The description to set on the build
            replace:      If True, replace existing description, if any, instead of appending to it
            separator:    A separator to insert between any existing description and the new :py:obj:`description` if :py:obj:`replace` is False.
            build_url:    The URL of the Jenkins job build.
            job_name:     Name of the Jenkins job
            build_number: The build number for which to set the description
        """

        self.poll()
        build_url = self.get_build_url(build_url, job_name, build_number)
        try:
            self._set_description(description, build_url, replace, separator)
        except ResourceNotFound as ex:
            raise ValueError(f"Build not found {repr(build_url)}", ex) from ex


class ApiJob():
    def __init__(self, jenkins, dct, name, has_children):
        self.jenkins = jenkins
        self.dct = dct.copy()
        self.name = name
        self.has_children = has_children
        self.public_uri = self.jenkins._public_job_url(self.name)  # pylint: disable=protected-access
        self.old_build_number = None
        self._invocations = {}
        self.queued_why = None

        self._path = "/job/" + self.name

        for prop in self.dct.get("property", []):
            if prop and prop.get('parameterDefinitions'):
                self._build_trigger_path = self._path + "/buildWithParameters"
                return

        for action in self.dct.get("action", []):
            if action and action.get('parameterDefinitions'):
                self._build_trigger_path = self._path + "/buildWithParameters"
                return

        self._build_trigger_path = self._path + "/build"

    def invoke(self, securitytoken, build_params, cause, description):
        try:
            if cause:
                build_params = build_params or {}
                build_params['cause'] = cause
            headers = _CT_URL_ENC if build_params else None
            params = {}
            if securitytoken:
                params['token'] = securitytoken
            response = self.jenkins.post(self._build_trigger_path, headers=headers, payload=build_params, **params)
        except ResourceNotFound as ex:
            raise UnknownJobException(self.jenkins._public_job_url(self.name), ex) from ex  # pylint: disable=protected-access

        # Make location relative
        location = urllib.parse.urlsplit(response.headers['location']).path
        if self.jenkins.jenkins_prefix:
            location = os.path.relpath(location, self.jenkins.jenkins_prefix)

        old_inv = self._invocations.get(location)
        if old_inv:
            old_inv.build_number = _SUPERSEDED_PSEUDO_BUILD_NUM
        inv = self.jenkins.invocation_class(self, location, description)
        self._invocations[location] = inv
        # print("invoke:", location, inv)
        return inv

    def poll(self):
        for invocation in self._invocations.values():
            # print("invocation:", invocation)
            if not invocation.build_number:
                query = "executable[number],why,*"
                dct = self.jenkins.get_json(invocation.queued_item_path, tree=query)
                # print("dct:", dct)

                executable = dct.get('executable')
                if executable:
                    invocation.build_number = executable['number']
                    invocation.queued_why = None
                    invocation.set_description()
                else:
                    try:
                        invocation.queued_why = dct['why']
                        # If we still have invocations in the queue, wait until next poll to query again
                        break
                    except KeyError:
                        # 'scans' have no 'why'
                        pass

    def job_status(self):
        """Result, progress and latest buildnumber info for the JOB, NOT the invocation

        Return (result, progress_info, latest_build_number) (str, str, int or None):
            If there is no finished build, result will be BuildResult.UNKNOWN and latest_build_number will be None
        """
        progress = None

        if not self.has_children:
            try:
                qi = self.dct['queueItem']
                if qi:
                    progress = Progress.QUEUED
                    self.queued_why = qi['why']
            except KeyError:
                # 'scans' have no 'queueItem'/'why' and they may not have children if scan has not run/finished
                pass

        dct = self.dct.get('lastBuild')
        if dct:
            self.old_build_number = dct['number']
            if dct["result"] is None:
                # Latest build dct does not have result, get it from builds list.
                query = "builds[number,result]"
                builds = self.jenkins.get_json(self._path, tree=query)
                for build_dct in builds['builds']:
                    if build_dct['number'] == self.old_build_number:
                        dct = build_dct
                        break

            result, latest_progress = _result_and_progress(dct)
            return (result, progress or latest_progress, self.old_build_number)

        return (BuildResult.UNKNOWN, progress or Progress.IDLE, None)

    def stop_all(self):
        # First remove pending builds from queue
        queue_item_ids = self.jenkins.queue_items.get(self.name) or []
        for qid in queue_item_ids:
            try:
                self.jenkins.post('/queue/cancelItem', id=repr(qid))
            except ResourceNotFound:
                # Job is no longer queued, so just ignore
                # NOTE: bug https://issues.jenkins-ci.org/browse/JENKINS-21311 also brings us here!
                pass

        # Abort running builds
        query = "builds[number,result]"
        dct = self.jenkins.get_json(self._path, tree=query)
        for build in dct['builds']:
            _result, progress = _result_and_progress(build)
            if progress != Progress.IDLE:
                build_number = build['number']
                try:
                    self.jenkins.post(self._path + '/' + repr(build_number) + '/stop')
                except ResourceNotFound:  # pragma: no cover
                    # Build was deleted, just ignore
                    pass

    def update_config(self, config_xml):
        self.jenkins.post(self._path + "/config.xml",
                          headers={'Content-Type': 'application/xml header;charset=utf-8'},
                          payload=config_xml.encode('utf-8'))

    def disable(self):
        try:
            self.jenkins.post(self._path + '/disable')
        except ResourceNotFound as ex:
            raise UnknownJobException(self.jenkins._public_job_url(self.name), ex) from ex  # pylint: disable=protected-access

    def enable(self):
        try:
            self.jenkins.post(self._path + '/enable')
        except ResourceNotFound as ex:
            raise UnknownJobException(self.jenkins._public_job_url(self.name), ex) from ex  # pylint: disable=protected-access

    def __repr__(self):
        return str(dict(name=self.name, dct=self.dct))


class Invocation(ApiInvocationMixin):
    def __init__(self, job, queued_item_path, description):
        self.job = job
        self.queued_item_path = queued_item_path
        self.description = description
        self.build_number = None
        self.queued_why = None

    def __repr__(self):
        return 'Invocation: ' + repr(self.queued_item_path) + ' ' + repr(self.build_number) + ' ' + repr(self.queued_why)

    def status(self):
        """Result and Progress info for the invocation

        Return (result, progress_info) (str, str):
            If the build has not started or has not finished running, result will be BuildResult.UNKNOWN
        """

        if self.build_number is None:
            return (BuildResult.UNKNOWN, Progress.QUEUED)

        if self.build_number == _SUPERSEDED_PSEUDO_BUILD_NUM:
            return (BuildResult.SUPERSEDED, Progress.IDLE)

        if self.build_number == _DEQUEUED_PSEUDO_BUILD_NUM:
            return (BuildResult.DEQUEUED, Progress.IDLE)

        # It seems that even after the executor has been assigned a number in the queue item, the lastBuild might not yet exist
        dct = self.job.dct.get('lastBuild')
        last_number = dct['number'] if dct else None
        if last_number is None:
            return (BuildResult.UNKNOWN, Progress.QUEUED)

        if last_number == self.build_number:
            return _result_and_progress(dct)

        if last_number < self.build_number:
            # TODO: Why does this happen?
            pass  # pragma: no cover

        # Latest build is not ours, get the correct build
        query = "builds[number,result]"
        dct = self.job.jenkins.get_json(self.job._path, tree=query)
        for build in dct['builds']:
            if build['number'] == self.build_number:
                return _result_and_progress(build)

        raise Exception("Build deleted while flow running? This may happen if you invoke more builds than the job is configured to keep. " + repr(self))

    def set_description(self):
        """Sets the build description"""
        if not self.description:
            return

        build_url = self.job._path + '/' + repr(self.build_number)
        try:
            self.job.jenkins._set_description(self.description, build_url)
        except ResourceNotFound as ex:
            raise Exception("Build deleted while flow running? " + repr(build_url), ex) from ex

    def stop(self, dequeue):
        try:
            if self.build_number is not None and self.build_number >= 0 and not dequeue:
                # Job has started
                self.job.jenkins.post(self.job._path + '/' + repr(self.build_number) + '/stop')
                return

            if self.build_number is None and dequeue:
                # Job is queued
                qid = self.queued_item_path.strip('/').split('/')[2]
                self.job.jenkins.post('/queue/cancelItem', id=qid)
                self.build_number = _DEQUEUED_PSEUDO_BUILD_NUM
        except ResourceNotFound:  # pragma: no cover
            # Job is no longer queued or running, except that it may have just changed from queued to running
            # We leave it up to the flow logic to handle that
            # NOTE: bug https://issues.jenkins-ci.org/browse/JENKINS-21311 also brings us here!
            pass



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, re, tempfile
from functools import partial
from pathlib import Path

from .lines_in import lines_in as generic_lines_in
from .cfg import ApiType
from .cfg.dirs import flow_graph_root_dir


_log_file_prefix = tempfile.gettempdir() + '/jenkinsflow/'


def _script_api_log_file(job_name):
    return _log_file_prefix + job_name + '.log'


def console_url(api, job_name, num):
    if api.api_type == ApiType.SCRIPT:
        return _script_api_log_file(job_name)
    return 'http://x.x/job/' + job_name + '/' + (str(num) + "/console" if api.api_type == ApiType.MOCK else "")


def result_msg(api, job_name, num=None):
    if api.api_type == ApiType.SCRIPT:
        return repr(job_name) + " - build: " + _script_api_log_file(job_name)
    return repr(job_name) + " - build: http://x.x/job/" + job_name + "/" + (str(num) + "/console" if api.api_type == ApiType.MOCK and num else "")


def build_started_msg(api, job_name, num, invocation_number=0):
    return "Build started: " + repr(job_name) + ((' Invocation-' + repr(invocation_number)) if invocation_number else '') + " - " + console_url(api, job_name, num)


def kill_current_msg(api, job_name, num):
    return "Killing build: " + repr(job_name) + " - " + console_url(api, job_name, num)


def build_queued_msg(api, job_name, num):
    if api.api_type == ApiType.MOCK:
        queued_why = r"Why am I queued\?"
    else:
        queued_why = r"Build #[0-9]+ is already in progress \(ETA: ?([0-9.]+ sec|N/A)\)"
    return re.compile("^job: '" + job_name + "' Status QUEUED - " + queued_why)


_HTTP_RE = re.compile(r'https?://.*?/job/([^/" ]*)(/?)')
def replace_host_port(api_type, contains_url):
    if api_type != ApiType.SCRIPT:
        return _HTTP_RE.sub(r'http://x.x/job/\1\2', contains_url)

    return _HTTP_RE.sub(r'/tmp/jenkinsflow-test/job/\1.py', contains_url)


def lines_in(api_type, text, *expected_lines):
    """Assert that `*expected_lines` occur in order in the lines of `text`.

    Args:
        *expected_lines (str or RegexObject (hasattr `match`)): For each `expected_line` in expected_lines:
            If `expected_line` has a match method it is called and must return True for a line in `text`.
            Otherwise, if the `expected_line` starts with '^', a line in `text` must start with `expected_line[1:]`
            Otherwise `expected line` must simply occur in a line in `text`
    """

    assert isinstance(api_type, ApiType)
    assert text
    assert expected_lines

    return generic_lines_in(text, partial(replace_host_port, api_type), *expected_lines)


def flow_graph_dir(flow_name: str, api_type: ApiType) -> Path:
    """Control which directory to put flow graph in.

    Put the generated graph in the workspace root if running from Jenkins.
    If running from commandline put it under config.flow_graph_root_dir/flow_name

    Return: dir-name
    """
    return Path.cwd() if os.environ.get('JOB_NAME') else Path(flow_graph_root_dir, flow_name + "-" + api_type.name.lower())



import os, socket

from . import dirs
from . import ApiType


class Urls():
    default_direct_url = "http://localhost:8080"
    proxied_public_url = "http://myproxy.mydom/jenkins"
    _cli_jar_path = 'jnlpJars/jenkins-cli.jar'

    def __init__(self, direct_url, script_dir):
        self._direct_url = direct_url.rstrip('/') if direct_url else self.default_direct_url
        self._script_dir = script_dir.rstrip('/') if script_dir else dirs.job_script_dir

    def direct_url(self, api_type):
        if api_type == ApiType.SCRIPT:
            return self._script_dir

        return self._direct_url

    def public_url(self, api_type):
        if api_type == ApiType.SCRIPT:
            return self._script_dir

        purl = os.environ.get('JENKINS_URL')
        return "http://" + socket.getfqdn() + ':' + repr(8080) + '/' if purl is None else purl

    def direct_cli_url(self, api_type):
        if api_type == ApiType.SCRIPT:
            return self._script_dir

        return self.direct_url(api_type) + '/' + self._cli_jar_path

    def public_cli_url(self, api_type):
        if api_type == ApiType.SCRIPT:
            return self._script_dir

        return self.public_url(api_type) + self._cli_jar_path

    def proxied_public_cli_url(self, api_type):
        # Not required to be a real url
        if api_type == ApiType.SCRIPT:
            return self._script_dir

        return self.proxied_public_url + '/' + self._cli_jar_path



# Copyright (c) 2024 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from datetime import datetime

import pytest

from jenkinsflow.flow import serial


from .framework.cfg import ApiType
from .framework import api_select


@pytest.mark.apis(ApiType.JENKINS)
@pytest.mark.parametrize("org_name", ["codeberg-org-jenkinsflow-test", "github-org-jenkinsflow-test"])
def test_org_folder_build_branch(api_type, org_name):
    with api_select.api(__file__, api_type, existing_jobs=True) as api:
        api.flow_job()
        api.job(f"{org_name}/jenkinsflow-org-job-test/main", max_fails=0, expect_invocations=1, expect_order=1)

        with serial(api, timeout=20, report_interval=1) as ctrl1:
            ctrl1.invoke(f"{org_name}/jenkinsflow-org-job-test/main")


@pytest.mark.apis(ApiType.JENKINS)
@pytest.mark.parametrize("org_name", ["codeberg-org-jenkinsflow-test", "github-org-jenkinsflow-test"])
def test_org_folder_build_slow_branch(api_type, org_name):
    with api_select.api(__file__, api_type, existing_jobs=True) as api:
        api.flow_job()
        api.job(f"{org_name}/jenkinsflow-org-job-test/slow", max_fails=0, expect_invocations=1, expect_order=1)

        before = datetime.now()
        with serial(api, timeout=70, report_interval=1) as ctrl1:
            ctrl1.invoke(f"{org_name}/jenkinsflow-org-job-test/slow")

        # The job sleeps 10s
        assert (datetime.now() - before).total_seconds() >= 10


@pytest.mark.apis(ApiType.JENKINS)
@pytest.mark.parametrize("org_name", ["codeberg-org-jenkinsflow-test", "github-org-jenkinsflow-test"])
def test_org_folder_build_with_parameters(api_type, org_name):
    with api_select.api(__file__, api_type, existing_jobs=True) as api:
        api.flow_job()
        api.job(f"{org_name}/jenkinsflow-org-job-test/parameters", max_fails=0, expect_invocations=1, expect_order=1)

        with serial(api, timeout=70, report_interval=1) as ctrl1:
            ctrl1.invoke(f"{org_name}/jenkinsflow-org-job-test/parameters", GREETING="Hi")


@pytest.mark.apis(ApiType.JENKINS)
@pytest.mark.parametrize("org_name", ["codeberg-org-jenkinsflow-test", "github-org-jenkinsflow-test"])
def test_org_folder_scan_organization(api_type, org_name):
    with api_select.api(__file__, api_type, existing_jobs=True) as api:
        api.flow_job()
        api.job(f"{org_name}", max_fails=0, expect_invocations=1, expect_order=1)

        # Scan time depends on the GitHub organization!
        with serial(api, timeout=10, report_interval=1) as ctrl1:
            ctrl1.invoke(f"{org_name}", assume_finished_after=5)


@pytest.mark.apis(ApiType.JENKINS)
@pytest.mark.parametrize("org_name", ["codeberg-org-jenkinsflow-test", "github-org-jenkinsflow-test"])
def test_org_folder_scan_repo(api_type, org_name):
    with api_select.api(__file__, api_type, existing_jobs=True) as api:
        api.flow_job()
        api.job(f"{org_name}/jenkinsflow-org-job-test", max_fails=0, expect_invocations=1, expect_order=1)

        # Repo scan time depends on the GitHub repo, it does not take too long, but can be queued for a while!
        with serial(api, timeout=10, report_interval=1) as ctrl1:
            ctrl1.invoke(f"{org_name}/jenkinsflow-org-job-test", assume_finished_after=5)



#!/usr/bin/env python3

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial

import lib.demo_security as security


def main(api):
    with serial(api, timeout=70, report_interval=3, job_name_prefix='jenkinsflow_demo__prefix__') as ctrl1:
        ctrl1.invoke('quick1')

        for index in 1, 2, 3:
            with ctrl1.serial(timeout=20, report_interval=3, job_name_prefix='x_') as ctrl2:
                ctrl2.invoke('quick2-' + str(index))

        ctrl1.invoke('quick3')

        with ctrl1.parallel(timeout=40, report_interval=3, job_name_prefix='y_') as ctrl2:
            with ctrl2.serial(timeout=40, report_interval=3, job_name_prefix='z_') as ctrl3:
                ctrl3.invoke('quick4')

            ctrl2.invoke('quick5')


if __name__ == '__main__':
    # Note: This flow uses username/password instead of securitytoken, to demonstrate that feature, it could have used securitytoken.
    # See demo_security.py
    import os
    from jenkinsflow.jenkins_api import JenkinsApi
    jenkins = JenkinsApi(os.environ.get('JENKINS_URL') or "http://localhost:8080", username=security.username, password=security.password)
    main(jenkins)



class JobLoad():
    def __init__(self, load_jobs: bool, delete_jobs: bool):
        self.load_jobs = load_jobs
        self.delete_jobs = delete_jobs

    def skip_job_delete(self):
        return not self.delete_jobs

    def skip_job_load(self):
        return not self.load_jobs

    def skip_job_load_sh_export_str(self):
        return 'export ' + JobLoad.skip_job_load_env_var_name + '=true'



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial

import pytest

from .framework import api_select
from .framework.cfg import ApiType


@pytest.mark.apis(ApiType.JENKINS)
def test_multiple_apis_mixed(api_type, options):
    _params = (('password', '', 'Some password'), ('s1', '', 'Some string argument'))

    if ApiType.SCRIPT not in options.apis:
        pytest.skip(f"Requires both {ApiType.JENKINS} and {ApiType.SCRIPT}")

    with api_select.api(__file__, api_type) as jenkins_api1:
        jenkins_api1.flow_job()

        jenkins_api1.job('job-1', max_fails=0, expect_invocations=1, expect_order=1, params=_params)
        jenkins_api1.job('job-3', max_fails=0, expect_invocations=1, expect_order=3)
        jenkins_api1.job('job-5', max_fails=0, expect_invocations=1, expect_order=3)
        jenkins_api1.job('job-7', max_fails=0, expect_invocations=1, expect_order=4, serial=True)

        with api_select.api(__file__, ApiType.SCRIPT) as script_api1:
            script_api1.job('job-2', max_fails=0, expect_invocations=1, expect_order=2, serial=True)
            script_api1.job('job-4', max_fails=0, expect_invocations=1, expect_order=3)
            script_api1.job('job-6', max_fails=0, expect_invocations=1, expect_order=3, params=_params, serial=True)

            with api_select.api(__file__, ApiType.JENKINS) as jenkins_api2:
                jenkins_api2.job('job-8', max_fails=0, expect_invocations=1, expect_order=3, serial=True)
                jenkins_api2.job('job-9', max_fails=0, expect_invocations=1, expect_order=3)

                # Flow
                with serial(jenkins_api1, timeout=70, job_name_prefix=jenkins_api1.job_name_prefix, report_interval=1) as ctrl1:
                    ctrl1.invoke('job-1', password='a', s1='b')
                    with ctrl1.serial(api=script_api1) as sc_ctrl:
                        sc_ctrl.invoke('job-2')

                    with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
                        with ctrl2.serial(timeout=40, report_interval=3) as ctrl3a:
                            ctrl3a.invoke('job-3')

                        with ctrl2.serial(api=script_api1, timeout=40, report_interval=3) as ctrl3a:
                            ctrl3a.invoke('job-6', password='a', s1='b')

                        with ctrl2.parallel(timeout=40, report_interval=3) as ctrl3b:
                            with ctrl3b.serial(api=script_api1, report_interval=0.5) as ctrl4:
                                ctrl4.invoke('job-4')
                            ctrl3b.invoke('job-5')

                        with ctrl2.parallel(api=jenkins_api2, timeout=40) as ctrl3c:
                            ctrl3c.invoke('job-8')
                            ctrl3c.invoke('job-9')

                    ctrl1.invoke('job-7')



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from ...framework import api_select


def create_jobs(api_type):
    api = api_select.api(__file__, api_type)
    def job(name, exec_time, max_fails=0, expect_order=0, expect_invocations=1, params=None):
        api.job(name, exec_time=exec_time, max_fails=max_fails, expect_invocations=expect_invocations, expect_order=expect_order, params=params)

    api.flow_job()
    job('wait1-1', 1, 0, 1)
    job('wait2', 2, 0, 2)
    job('wait5-2a', 5, 0, 2)
    job('quick_fail-2', 0.1, 1, 2, params=(('s1', 'WORLD', 'desc'), ('c1', ('maybe', 'certain'), 'desc')))
    job('wait5-2b', 5, 0, 2)
    job('wait5-2c', 5, 0, 2)
    job('quick_fail-1', 0.5, 1, 2, params=(('s1', 'WORLD', 'desc'), ('c1', ('to', 'be', 'or', 'not'), 'desc')))
    job('quick', 0.5, 0, 2, expect_invocations=0, params=(('s1', 'WORLD', 'desc'), ('c1', ('maybe', 'certain'), 'desc')))
    job('wait1-2', 1, 0, 3, expect_invocations=0)
    return api


if __name__ == '__main__':
    create_jobs(api_select.ApiType.JENKINS)



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from ...framework import api_select


def create_jobs(api_type):
    api = api_select.api(__file__, api_type)
    def job(name, expect_order):
        api.job(name, exec_time=0.5, max_fails=0, expect_invocations=1, expect_order=expect_order, params=None)

    api.flow_job()
    job('quick1', 1)
    index = 0
    for index in 1, 2, 3:
        job('x_quick2-' + str(index), 1+index)
    job('quick3', 2+index)
    job('y_z_quick4', 3+index)
    job('y_quick5', 3+index)
    return api


if __name__ == '__main__':
    create_jobs(api_select.ApiType.JENKINS)



#!/usr/bin/env python3

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import click


from .set_build_description import set_build_description


@click.group()
def cli():
    """Commandline utilities for jenkinsflow"""


cli.add_command(set_build_description)


def main():
    cli(auto_envvar_prefix='JENKINSFLOW')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from jenkinsflow.flow import serial

import lib.get_jenkins_api


def main(api, securitytoken):
    with serial(api, timeout=70, securitytoken=securitytoken, job_name_prefix='jenkinsflow_demo__errors__', report_interval=3) as ctrl1:
        ctrl1.invoke('wait1-1')

        with ctrl1.parallel(timeout=40, report_interval=3) as ctrl2:
            with ctrl2.serial(timeout=40, report_interval=3) as ctrl3a:
                ctrl3a.invoke('wait2')
                ctrl3a.invoke('quick_fail-1', force_result='FAILURE', password='HELLO', s1='WORLD', c1='be')

                # Never invoked
                ctrl3a.invoke('quick', password='HELLO', s1='WORLD', c1='maybe')

            with ctrl2.parallel(timeout=40, report_interval=3) as ctrl3b:
                ctrl3b.invoke('wait5-2a')
                ctrl3b.invoke('quick_fail-2', force_result='FAILURE', password='HELLO', s1='WORLD', c1='maybe')
                ctrl3b.invoke('wait5-2b')
                ctrl3b.invoke('wait5-2c')

        # Never invoked
        ctrl1.invoke('wait1-2')


if __name__ == '__main__':
    main(*lib.get_jenkins_api.get_jenkins_api())



import os.path

import requests

from .api_base import AuthError, ClientError


def _join_url(start, end):
    return os.path.join(start.rstrip('/'), end.lstrip('/'))


class ResourceNotFound(Exception):
    pass


class RequestsRestApi():
    def __init__(self, direct_uri, username, password):
        super().__init__()
        self.session = requests.Session()
        if username or password:
            self.session.auth = requests.auth.HTTPBasicAuth(username, password.encode("utf-8"))
        self.direct_uri = direct_uri

    @staticmethod
    def _check_response(response, good_responses=(200,)):
        if response.status_code in good_responses:
            return response

        try:
            response.raise_for_status()
        except Exception as ex:
            if response.status_code == 404:
                raise ResourceNotFound(ex) from ex
            if response.status_code == 401:
                raise AuthError(ex) from ex
            if response.status_code == 403:
                raise ClientError(ex) from ex
            if response.status_code == 500:
                # TODO: Workaround for https://issues.jenkins.io/browse/JENKINS-63845
                # 500 Server Error: Server Error for url: http://localhost:8080/queue/cancelItem?id=14406
                if "Server Error for url:" in str(ex) and "queue/cancelItem" in str(ex):
                    raise ResourceNotFound(ex) from ex
            raise

        # TODO: This is dubious, maybe we should raise here instead.
        return response

    def _get(self, url, params):
        try:
            return self._check_response(self.session.get(_join_url(self.direct_uri, url), params=params))
        except requests.ConnectionError as ex:
            raise ConnectionError(ex) from ex

    def get_content(self, url, **params):
        return self._get(url, params=params).content

    def get_json(self, url="", **params):
        return self._get(_join_url(url, "api/json"), params=params).json()

    def post(self, url, payload=None, headers=None, allow_redirects=False, **params):
        response = self.session.post(_join_url(self.direct_uri, url), headers=headers, data=payload, allow_redirects=allow_redirects, params=params)
        return self._check_response(response, (200, 201))

    def headers(self, allow_redirects=True):
        return self._check_response(self.session.head(self.direct_uri, allow_redirects=allow_redirects)).headers



# Copyright (c) 2012 - 2024 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys
import os


def env_base_url():
    """Determines base_url when run from within a Jenkins job

    Returns:
        base_url (str): If JENKINS_URL is set, i.e. run from within job.

    Raises:
        Exception if JENKINS_URL is not set, i.e. NOT run from within job.
    """

    base_url = os.environ.get('JENKINS_URL')
    if base_url:
        return base_url.rstrip('/')

    if base_url is None:
        raise ValueError(
            "Could not get env variable JENKINS_URL. You must set 'Jenkins Location' in Jenkins setup for JENKINS_URL to be exported.")


def base_url_and_api(direct_url):
    try:
        base_url = direct_url.rstrip('/') if direct_url else env_base_url()
    except Exception:
        print("*** ERROR: You must specify 'direct-url' if not running from Jenkins job", file=sys.stderr)
        raise

    if base_url.startswith('http:') or base_url.startswith('https:'):
        # Using jenkins_api
        from ..jenkins_api import JenkinsApi as api
    else:
        # Using script_api
        from ..script_api import  ScriptApi as api

    return base_url, api



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from pathlib import Path

import pytest

from jenkinsflow.flow import serial, FailedChildJobException

from .framework import api_select
from .framework.cfg import ApiType


_HERE = Path(__file__).resolve().parent
_FRAMEWORK_DIR = _HERE/"framework"


@pytest.mark.apis(ApiType.JENKINS)
def test_early_set_result(api_type):
    """Make sure we keep checking until we have the final result, not an intermediate one."""
    with api_select.api(__file__, api_type) as api:
        api.job('wait10', max_fails=1, expect_invocations=1, expect_order=1, exec_time=10, serial=True, job_xml_template=_FRAMEWORK_DIR/"pipeline_job.xml.tenjin")

        with pytest.raises(FailedChildJobException):
            with serial(api, timeout=40, job_name_prefix=api.job_name_prefix, report_interval=1) as ctrl:
                ctrl.invoke('wait10')



# Copyright (c) 2019 - 2024 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from .ordered_enum import OrderedEnum


class Propagation(OrderedEnum):
    NORMAL = 0
    FAILURE_TO_UNSTABLE = 1
    UNCHECKED = 2



# Copyright (c) 2012 - 2015 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os, sys, re

from jenkinsflow.unbuffered import UnBuffered

from ..conftest import get_cfg

from .cfg import ApiType, AllCfg
from .cfg.speedup import speedup


sys.stdout = UnBuffered(sys.stdout)

_file_name_subst = re.compile(r'(_jobs|_test)?\.py')


def api(file_name, api_type, login=None, fixed_prefix=None, url_or_dir=None, fake_public_uri=None, invocation_class=None,
        username=None, password=None, *, options: AllCfg = None, existing_jobs: bool = False):
    """Factory to create either Mock or Wrap api"""
    options = options or get_cfg()
    base_name = os.path.basename(file_name).replace('.pyc', '.py')
    job_name_prefix = _file_name_subst.sub('', base_name)
    func_name = None
    func_num_params = 0
    if existing_jobs:
        job_name_prefix = ""
        file_name = base_name
    elif fixed_prefix:
        job_name_prefix = fixed_prefix
        file_name = base_name
    elif '_test' in file_name:
        func_name = sys._getframe().f_back.f_code.co_name  # pylint: disable=protected-access
        func_num_params = sys._getframe().f_back.f_code.co_argcount  # pylint: disable=protected-access
        file_name = base_name
        func_name = func_name.replace('test_', '')
        assert func_name[0:len(job_name_prefix)] == job_name_prefix, \
            "Naming standard not followed: " + repr('test_' + func_name) + " defined in file: " + repr(base_name) + " should be 'test_" + job_name_prefix + "_<sub test>'"
        job_name_prefix = 'jenkinsflow_test__' + func_name + '__'
    else:
        job_name_prefix = 'jenkinsflow_demo__' + job_name_prefix + '__'
        file_name = base_name.replace('_jobs', '')

    print()
    print("--- Preparing api for ", repr(job_name_prefix), "---")

    if api_type == 0:
        # Invocation from actual Jenkins flow job calls with api_type == 0
        api_type = ApiType.JENKINS
    print('Using:', api_type)

    url_or_dir = url_or_dir or options.urls.direct_url(api_type)
    reload_jobs = not options.job_load.skip_job_load() and not fixed_prefix and not existing_jobs
    pre_delete_jobs = not options.job_load.skip_job_delete()

    from .cfg import jenkins_security
    if login is None:
        login = jenkins_security.default_use_login

    if password is not None or username is not None:
        assert password is not None and username is not None
        login = True

    if username is None:
        assert password is None
        username = jenkins_security.username
        password = jenkins_security.password

    if api_type == ApiType.JENKINS:
        from .api_wrapper import JenkinsTestWrapperApi
        return JenkinsTestWrapperApi(
            file_name=file_name,
            func_name=func_name,
            func_num_params=func_num_params,
            job_name_prefix=job_name_prefix,
            reload_jobs=reload_jobs,
            pre_delete_jobs=pre_delete_jobs,
            direct_url=url_or_dir,
            fake_public_uri=fake_public_uri,
            username=username,
            password=password,
            securitytoken=jenkins_security.securitytoken,
            login=login,
            invocation_class=invocation_class,
            python_executable=os.environ["JEKINSFLOW_TEST_JENKINS_API_PYTHON_EXECUTABLE"],
            existing_jobs=existing_jobs)

    if api_type == ApiType.SCRIPT:
        from .api_wrapper import ScriptTestWrapperApi
        return ScriptTestWrapperApi(
            file_name=file_name,
            func_name=func_name,
            func_num_params=func_num_params,
            job_name_prefix=job_name_prefix,
            reload_jobs=reload_jobs,
            pre_delete_jobs=pre_delete_jobs,
            direct_url=url_or_dir,
            fake_public_uri=fake_public_uri,
            username=username,
            password=password,
            securitytoken=jenkins_security.securitytoken,
            login=login,
            invocation_class=invocation_class)

    if api_type == ApiType.MOCK:
        from .mock_api import MockApi
        return MockApi(
            job_name_prefix=job_name_prefix,
            speedup=speedup(),
            public_uri=options.urls.direct_url(api_type),
            python_executable=sys.executable)

    raise Exception(f"Unhandled api_type: {repr(api_type)} - {api_type.__class__.__module__} was compared to {ApiType.MOCK.__class__.__module__}")



"""Configuration file for 'pytest'"""

import sys
import os
import re
import errno
from pathlib import Path
import shutil

import pytest
from pytest import fixture
from click.testing import CliRunner

from .framework import pytest_options
from .framework.cfg import ApiType, opts_to_test_cfg

# Note: You can't (indirectly) import stuff from jenkinsflow here, it messes up the coverage

_HERE = Path(__file__).absolute().parent
_DEMO_DIR = (_HERE/"../demo").resolve()
sys.path.append(str(_DEMO_DIR))

_OUT_DIRS = {}

def _test_key_shortener(key_prefix, key_suffix):
    prefix = key_prefix.replace('test.', '').replace('_test', '')
    suffix = key_suffix.replace(prefix, '').replace('test_', '').strip('_')
    outd = prefix + '.' + suffix
    args = (key_prefix, key_suffix)
    assert _OUT_DIRS.setdefault(outd, args) == args, \
        f"Out dir name '{outd}' reused! Previous from {_OUT_DIRS[outd]}, now  {args}. Test is not following namimg convention."
    return outd


def _test_node_shortener(request):
    """Shorten test node name while still keeping it unique"""
    return _test_key_shortener(request.node.module.__name__, request.node.name.split('[')[0])


@fixture(name="out_dir")
def _fixture_out_dir(request):
    """Create unique top level test directory for a test."""

    out_dir = _HERE/'out'/_test_node_shortener(request)

    try:
        shutil.rmtree(out_dir)
    except OSError as ex:
        if ex.errno != errno.ENOENT:
            raise

    return out_dir

# Add you configuration, e.g. fixtures here.


def pytest_addoption(parser):
    """pytest hook"""
    pytest_options.add_options(parser)


def pytest_configure(config):
    """pytest hook"""
    # Register api  marker
    config.addinivalue_line("markers", "apis(*ApiType): mark test to run only when using specified apis")
    config.addinivalue_line("markers", "not_apis(*ApiType): mark test NOT to run when using specified apis")

    config.cuctom_cfg = opts_to_test_cfg(
        config.getoption(pytest_options.OPT_DIRECT_URL),
        config.getoption(pytest_options.OPT_JOB_LOAD),
        config.getoption(pytest_options.OPT_JOB_DELETE),
        config.getoption(pytest_options.OPT_MOCK_SPEEDUP),
        config.getoption(pytest_options.OPT_API),
    )
    pytest._CUSTOM_TEST_CFG = config.cuctom_cfg


def get_cfg():
    return pytest._CUSTOM_TEST_CFG


def pytest_collection_modifyitems(items: list[pytest.Item], config) -> None:
    """pytest hook"""
    selected_api_types = config.cuctom_cfg.apis
    item_api_type_regex = re.compile(r'.*\[ApiType\.(.*)\]')
    remaining = []
    deselected = []

    def filter_items_by_api_type(item):
        if "api_type" not in item.fixturenames:
            for om in item.own_markers:
                if om.name in ("apis", "not_apis"):
                    location = ':'.join(str(place) for place in item.location)
                    raise Exception(f"{location}: Error: A test using the 'apis' or 'not_apis' marker must also use the 'api_type' fixture.")
            remaining.append(item)
            return

        # We must have an ApiType in name now, since api_type fixture was used
        api_type_name, _, _ = (item_api_type_regex.match(item.name).groups()[0]).partition("-")  # TODO, this cannot be the way to get api_type
        current_item_api = ApiType[api_type_name]
        if current_item_api not in selected_api_types:
            print(f"{':'.join(str(place) for place in item.location)} DESELECTED ({current_item_api} not in {selected_api_types})")
            deselected.append(item)
            return

        remaining.append(item)


    print()
    for item in items:
        filter_items_by_api_type(item)

    if deselected:
        config.hook.pytest_deselected(items=deselected)
        items[:] = remaining


@pytest.fixture(scope="session")
def options(pytestconfig):
    """Access to test configuration objects."""
    return pytestconfig.cuctom_cfg


@pytest.fixture(params=list(ApiType))
def api_type(request, options):
    """ApiType fixture"""
    selected_api_type = request.param
    assert isinstance(selected_api_type, ApiType)

    for apimarker in request.node.iter_markers("apis"):
        apis = apimarker.args
        for allowed_api_type in apis:
            assert isinstance(allowed_api_type, ApiType)

        if selected_api_type not in apis:
            pytest.skip(f"only for {apis} APIs, current {selected_api_type}")
            return selected_api_type

    for not_apimarker in request.node.iter_markers("not_apis"):
        not_apis = not_apimarker.args
        for not_api_type in not_apis:
            assert isinstance(not_api_type, ApiType)

        if selected_api_type in not_apis:
            pytest.skip(f"not for {not_apis} APIs, current {selected_api_type}")
            return selected_api_type

    return selected_api_type


def _set_env_fixture(var_name, value, request):
    """
    Ensure env var_name is set to the value <value>
    Set back to original value, if any, or unset it, after test.
    """
    has_var = os.environ.get(var_name)
    os.environ[var_name] = value
    if not has_var:
        def fin():
            del os.environ[var_name]
    else:
        def fin():
            os.environ[var_name] = has_var
    request.addfinalizer(fin)


def _set_jenkins_url_env_fixture(not_set_value, request):
    if os.environ.get('JENKINS_URL'):
        _set_env_fixture('JENKINS_URL', not_set_value, request)


def _set_env_if_not_set_fixture(var_name, not_set_value, request):
    """
    Ensure env var_name is set to the value 'not_set_value' IFF it was not already set.
    Unset it after test.
    """
    has_var = os.environ.get(var_name)
    if not has_var:
        os.environ[var_name] = not_set_value
        def fin():
            del os.environ[var_name]
        request.addfinalizer(fin)


def _set_jenkins_url_env_if_not_set_fixture(not_set_value, request):
    _set_env_if_not_set_fixture('JENKINS_URL', not_set_value, request)


def _unset_env_fixture(var_name, request):
    """
    Ensure env var_name is NOT set
    Set back to original value, if any, after test.
    """
    has_var = os.environ.get(var_name)
    if has_var:
        del os.environ[var_name]
        def fin():
            os.environ[var_name] = has_var
        request.addfinalizer(fin)


@fixture
def env_base_url(request, api_type, options):
    # Fake that we are running from inside jenkins job
    public_url = options.urls.public_url(api_type)
    _set_jenkins_url_env_if_not_set_fixture(public_url, request)
    return public_url


@fixture
def env_base_url_trailing_slash(request, api_type, options):
    _set_jenkins_url_env_if_not_set_fixture(options.urls.public_url(api_type) + '/', request)


@fixture
def env_base_url_trailing_slashes(request, api_type, options):
    _set_jenkins_url_env_if_not_set_fixture(options.urls.public_url(api_type) + '//', request)


@fixture
def env_no_base_url(request):
    # Make sure it looks as if we are we are running from outside jenkins job
    _unset_env_fixture('JENKINS_URL', request)


@fixture
def env_different_base_url(request, options):
    # Fake that we are running from inside jenkins job
    # This url is not used, but should simply be different fron direct_url used in test, to simulate proxied jenkins
    _set_jenkins_url_env_fixture(options.urls.proxied_public_url, request)


@fixture
def env_job_name(request):
    # Fake that we are running from inside jenkins job
    _set_env_if_not_set_fixture('JOB_NAME', 'hudelihuu', request)


@fixture
def env_build_number(request):
    # Fake that we are running from inside jenkins job
    _set_env_if_not_set_fixture('BUILD_NUMBER', '1', request)


@fixture(scope='function')
def cli_runner(request):
    return CliRunner()



"""nox https://nox.thea.codes/en/stable/ configuration"""

# Use nox >= 2023.4.22

import os
import sys
import argparse
from pathlib import Path
import subprocess
import shutil
from typing import Sequence

import nox


_HERE = Path(__file__).absolute().parent
_TEST_DIR = _HERE/"test"
_DEMO_DIR = _HERE/"demo"
_DOC_DIR = _HERE/"docs"

sys.path.extend((str(_HERE), str(_DEMO_DIR)))

from test.framework.cfg import ApiType, str_to_apis, dirs
from test.framework.pytest_options import add_options


# Locally we have nox handle the different versions, but in each travis run there is only a single python which can always be found as just 'python'
_PY_VERSIONS = ["3.13", "3.12", "3.11", "3.10"] if not os.environ.get("TRAVIS_PYTHON_VERSION") else ["python"]
_IS_CI = os.environ.get("CI", "false").lower() == "true"

nox.options.error_on_missing_interpreters = True
nox.options.envdir = dirs.pseudo_install_dir


def _cov_options_env(api_types: Sequence[ApiType], env: dict[str, str], fail_under: int = 0) -> Sequence[str]:
    """Setup coverage options.

    Return pytest coverage options, and env variables dict.
    """

    if not fail_under:
        if len(api_types) == 3:
            fail_under = 93.90
        elif ApiType.JENKINS in api_types:
            fail_under = 94.90
        elif ApiType.MOCK in api_types and ApiType.SCRIPT in api_types:
            fail_under = 86.97
        elif ApiType.MOCK in api_types:
            fail_under = 85.33
        elif ApiType.SCRIPT in api_types:
            fail_under = 81.66
        else:
            raise ValueError(f"Unknown coverage requirement for api type combination: {api_types}")

    # Set coverage exclude lines based on selected API types
    api_exclude_lines = []
    if api_types == [ApiType.SCRIPT]:
        # Parts of api_base not used in script_api (overridden methods)
        api_exclude_lines.append(r"return (self.job.public_uri + '/' + repr(self.build_number) + '/console')")

    # Set coverage exclude files based on selected API type
    api_exclude_files = []
    if ApiType.JENKINS not in api_types:
        api_exclude_files.append("jenkins_api.py")
    if ApiType.SCRIPT not in api_types:
        api_exclude_files.extend(["script_api.py", "logging_process.py"])

    env["COV_API_EXCLUDE_LINES"] = "\n".join(api_exclude_lines)
    env["COV_API_EXCLUDE_FILES"] = "\n".join(api_exclude_files)

    return ['--cov', '--cov-report=term-missing', f'--cov-fail-under={fail_under}', f'--cov-config={_TEST_DIR/".coveragerc"}']


def _pytest_parallel_options(parallel, api_types):
    args = []

    if api_types != [ApiType.MOCK]:
        # Note: 'forked' is required for the kill/abort_current test not to abort other tests
        args.append('--forked')

        # Note: parallel actually significantly slows down the test when only running mock
        if parallel:
            args.extend(['-n', '16'])

    return args


_TEST_AND_DEMOS_INSTALLED = False

def _test_and_demos_install(session):
    """Make test and demos available to Jenkins process."""
    global _TEST_AND_DEMOS_INSTALLED

    site_packages_dir = Path(session.bin, f"../lib/python{session.python}/site-packages/jenkinsflow").resolve()

    for src_dir_name in "demo", "test":
        tgt_dir = Path(dirs.test_tmp_dir, src_dir_name)
        if not _TEST_AND_DEMOS_INSTALLED:
            exclude = [f"--exclude={exc}" for exc in (".git", "*~", "'*.py[cod]'", "__pycache__", "*.cache")]
            subprocess.check_call(["rsync", "-a", "--delete", "--delete-excluded", *exclude, f"{_HERE}/{src_dir_name}/", f"{tgt_dir}/"])
            for root, _, files in os.walk(tgt_dir):
                os.chmod(root, 0o777)
                for fn in files:
                    os.chmod(Path(root, fn), 0o644)

        lnk = site_packages_dir/src_dir_name
        lnk.unlink(missing_ok=True)
        lnk.symlink_to(tgt_dir)

    _TEST_AND_DEMOS_INSTALLED = True


# @nox.session(python=_PY_VERSIONS, reuse_venv=True)
# def typecheck(session):
#     session.install("-e", ".", "mypy>=1.5.1")
#     session.run("mypy", str(_HERE/"src"))


@nox.session(python=_PY_VERSIONS[0], reuse_venv=True)
def pylint(session):
    session.install(".", "pylint>=3.3.1", "pylint-pytest>=1.1.8", "objproxies>=0.9.4")

    print("\nPylint src")
    session.run("pylint", "--fail-under", "8.5", str(_HERE/"src"))

    print("\nPylint test sources")
    disable_checks = "missing-module-docstring,missing-class-docstring,missing-function-docstring"
    disable_checks += ",multiple-imports,invalid-name,duplicate-code"

    test_wrapper_api_members = "api_type,get_job,get_name_for_query,job_loader_jenkins,job_name_prefix,job_xml_template,test_jobs,poll"
    generated_members = ",".join(f"test.framework.api_wrapper._TestWrapperApi.{member}" for member in test_wrapper_api_members.split(","))

    session.run(
        "pylint", "--fail-under", "9.2", "--variable-rgx", r"[a-z_][a-z0-9_]{1,30}$", "--disable", disable_checks,
        "--generated-member", generated_members,
        "--ignore", "jenkins_security.py,demos_test.py", str(_TEST_DIR))


@nox.session(python=_PY_VERSIONS, reuse_venv=True)
def unit(session):
    """
    Test jenkinsflow.
    Runs all tests mocked in hyperspeed, runs against Jenkins, using jenkins_api, and run script_api jobs.

    Normally jobs will be run in parallel, specifying 'job_delete' disables this.
    The default options assumes that re-loading without deletions generates correct job config.
    Tests that require jobs to be deleted/non-existing will delete the jobs, regardless of the 'job_delete' option.

    Will process some of the special pytest args:
        direct_url: Direct Jenkins URL. Must be different from the URL set in Jenkins (and preferably non proxied).
        apis: The apis totest, default all.
        job_delete: Delete and re-load jobs into Jenkins.
        job_load: Load jobs into Jenkins (skipping job load assumes all jobs already loaded and up to date).
    """

    session.install("--upgrade", ".", "-r", str(_TEST_DIR/"requirements.txt"))

    pytest_args = []
    if _IS_CI:
        pytest_args.append("-vvv")

    parser = argparse.ArgumentParser(description="Process pytest options")
    add_options(parser)
    parsed_args = parser.parse_known_args(session.posargs)
    # print("parsed_args:", parsed_args)
    special_args = parsed_args[0]
    unknown_pytest_args = parsed_args[1]

    apis = str_to_apis(special_args.api)
    # print("noxfile, apis:", apis)

    cov_file = _HERE/".coverage"
    if os.path.exists(cov_file):
        os.remove(cov_file)

    env = {}

    # If we have unknow pytest options we don't know what the coverage would be
    fail_under = 1 if unknown_pytest_args and (unknown_pytest_args[0] != "-n" or len(unknown_pytest_args) != 2) else 0
    cov_opts = _cov_options_env(apis, env, fail_under)
    # env["COVERAGE_DEBUG"] = "config,trace,pathmap"

    parallel = special_args.job_load or special_args.job_delete
    pytest_args.extend(_pytest_parallel_options(parallel, apis))
    pytest_args.extend(["--capture=sys", "--instafail", "-p", "no:warnings", "--failed-first"])
    pytest_args.extend(session.posargs)

    if apis != [ApiType.MOCK]:
        env["JEKINSFLOW_TEST_JENKINS_API_PYTHON_EXECUTABLE"] = f"{session.bin}/python"
        _test_and_demos_install(session)

    session.run("pytest", "--capture=sys", *cov_opts, *pytest_args, env=env)


@nox.session(python=_PY_VERSIONS[0], reuse_venv=True)
def build(session):
    if Path("dist").is_dir():
        shutil.rmtree("dist")
    session.install("build>=1.0.3", "twine>=4.0.2")
    session.run("python", "-m", "build")
    session.run("python", "-m", "twine", "check", "dist/*")


@nox.session(python=_PY_VERSIONS[0], reuse_venv=True)
def docs(session):
    session.install("--upgrade", ".", "-r", str(_DOC_DIR/"requirements.txt"))
    session.run("make", "-C", "docs/source", "html")


@nox.session(python=_PY_VERSIONS[0], reuse_venv=True)
def cli(session):
    session.install("--upgrade", ".")
    session.run("jenkinsflow", "set-build-description", "--help")



# Copyright (c) 2012 - 2024 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import os
from enum import Enum
import urllib.parse

from .ordered_enum import OrderedEnum


class BuildResult(OrderedEnum):
    FAILURE = 0
    ABORTED = 1
    UNSTABLE = 2
    SUCCESS = 3
    SUPERSEDED = 4
    DEQUEUED = 5
    UNKNOWN = 6


class Progress(Enum):
    RUNNING = 1
    QUEUED = 2
    IDLE = 3


class AuthError(Exception):
    pass


class ClientError(AuthError):
    pass


class UnknownJobException(Exception):
    def __init__(self, job_url, api_ex=None):
        super().__init__("Job not found: " + job_url + (", " + repr(api_ex) if api_ex is not None else ""))


class InvalidJobNameException(Exception):
    def __init__(self, job_name):
        super().__init__(f"Invalid job name '{job_name}'. Job name should be either '<name>' or '<github-organization>[/<repo>[/<branch>]]'.")


class BaseApiMixin():
    def get_build_url(self, build_url: str = None, job_name: str = None, build_number: int = None):
        """Get the build_url either from arguments or environment.

        Arguments are preferred if specified and 'build_url' is preferred over job_name and build_number.
        """

        if build_url is None and job_name is None and build_number is None:
            try:
                build_url = os.environ['BUILD_URL']
            except KeyError:
                pass

        if build_url is not None:
            # Make build_url relative
            build_url = urllib.parse.urlsplit(build_url).path
            if self.jenkins_prefix:
                return os.path.relpath(build_url, self.jenkins_prefix)
            return build_url

        job_name = job_name if job_name is not None else os.environ['JOB_NAME']
        build_number = build_number if build_number is not None else int(os.environ['BUILD_NUMBER'])
        return f"/job/{job_name}/{build_number}"

    @staticmethod
    def get_name_for_query(job_name_prefix: str, job_name: str) -> str:
        job_name = job_name.strip("/")
        parts = job_name.split("/")

        if len(parts) == 1:
            return job_name_prefix + job_name

        if len(parts) == 2:
            return parts[0] + "/job/" + job_name_prefix + parts[1]

        if len(parts) == 3:
            return parts[0] + "/job/" + job_name_prefix + parts[1] +  "/job/" + parts[2]

        raise InvalidJobNameException(job_name)


class ApiInvocationMixin():
    def console_url(self):
        return (self.job.public_uri + '/' + repr(self.build_number) + '/console') if self.build_number is not None else None



from enum import Enum


class ApiType(Enum):
    JENKINS = 0
    SCRIPT = 1
    MOCK = 2


def str_to_apis(api_option: str) -> list[ApiType]:
    return [ApiType[api.strip().upper()] for api in api_option.split(',')]


def apis_to_str(apis: list[ApiType]) -> str:
    return ','.join([api.name for api in apis])



# This file is for configuring security for the demo and test jobs
# See the basic.py demo for how to do this with your own flows

# Note:

# Securitytoken and username/password are two different ways of allowing api access to execute (build) a job.
# You will however need both authentication methods, since neither can be used for everything

# Securitytoken is SIGNIFICANTLY faster than username/password
# If securitytoken is set, it will be used for executing (building) the jobs
# Securitytoken MUST be used in order for the flow job to ammend the 'build cause' of jobs that it invokes, if username/password
# is used the cause will always be 'Invoked by <username>'

# Username and password are needed for creating jobs and for changing build status with the 'Propagation.FAILURE_TO_UNSTABLE' feature
# The ./test/test.py script will need to create Jenkins jobs which REQUIRES username/password

# In order for all tests to pass you must enable security, select a security realm and a select "Anyone can do anything" in "Authorization",
# "Logged-in users can do anything" + "Allow anonymous read access" or Matrix/Projext/Role based security.
# If you have enabled any security beyond "Anyone can do anything" you must set 'set_build_description_must_authenticate = True' below.

# If you are using Matrix/Projext/Role based security, then the test user defined below and the user 'jenkinsflow_authtest1' should have
# the following permissions:

#  Job:   Build Cancel Configure Create Delete Discover Read
#  Build: Update

# The 'Anonymous' user must have - Job: Discover
# The 'authenticated' group must have - Overall: Read, Job: Discover

# If using "Role-Based Strategy":
#  create an item-role 'jenkinsflow_test' with pattern 'jenkinsflow_.*' and assign the permissions, then add the test user to that group.
#  create a global role 'jenkinsflow_test' with Create permission, then add the test user to that group.
# Job creation does not work with a project role only, see: https://issues.jenkins-ci.org/browse/JENKINS-19934?attachmentViewMode=list

# Must be True unless "Anyone can do anything" or "Logged-in users can do anything" is selected in security settings.
default_use_login = True

set_build_description_must_authenticate = True

securitytoken = 'jenkinsflow_securitytoken'

username = 'jenkinsflow_jobrunner'
password = '123123'



# Copyright (c) 2012 - 2024 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import sys
from . import ApiType, str_to_apis, apis_to_str, Urls, JobLoad, dirs, speedup


class AllCfg():
    def __init__(self, urls: Urls, job_load: JobLoad, apis: list[ApiType]):
        self.urls = urls
        self.job_load = job_load
        self.apis = apis


def opts_to_test_cfg(direct_url: str, load_jobs: bool, delete_jobs: bool, mock_speedup: int, apis_str: str) -> AllCfg:
    urls = Urls(direct_url, dirs.job_script_dir)
    job_load = JobLoad(load_jobs, delete_jobs)

    speedup.select_speedup(mock_speedup)

    try:
        apis = str_to_apis(apis_str)
        print("APIs:", apis)
    except KeyError as ex:
        print(ex, file=sys.stderr)
        print(f"'{ex}' cannot be converted to ApiType. Should be one or more of: {','.join([at.name for at in list(ApiType)])}")
        raise

    return AllCfg(urls, job_load, apis)


def opt_strs_to_test_cfg(direct_url: str, load_jobs: str, delete_jobs: str, mock_speedup: str, apis_str: str) -> AllCfg:
    return opts_to_test_cfg(
        direct_url = direct_url,
        load_jobs = load_jobs.lower() == "true",
        delete_jobs = delete_jobs.lower() == "true",
        mock_speedup = int(mock_speedup),
        apis_str = apis_str)


def test_cfg_to_opt_strs(test_cfg: AllCfg, api_type: ApiType) -> tuple[str, str, str, str, str]:
    return (
        test_cfg.urls.direct_url(api_type),
        str(test_cfg.job_load.load_jobs).lower(),
        str(test_cfg.job_load.delete_jobs).lower(),
        str(speedup.speedup()),
        apis_to_str(test_cfg.apis),
    )



# Copyright (c) 2012 - 2024 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

from ...framework import api_select


def create_jobs(api_type):
    component_groups = {
        'g1': range(1),
        'g2': range(2),
        'g3': range(2),
    }

    api = api_select.api(__file__, api_type)
    def job(name, expect_order, params=None):
        api.job(name, exec_time=0.5, max_fails=0, expect_invocations=1, expect_order=expect_order, params=params)

    api.flow_job()
    job('prepare', 1)
    for gname, group in component_groups.items():
        for component in group:
            job('deploy_component_' + gname + '_' + str(component), 2)
    job('report_deploy', 3)
    job('prepare_tests', 3)
    job('test_ui', 4)
    job('test_x', 4)
    for gname, group in component_groups.items():
        for component in group:
            job('test_component_' + gname + '_' + str(component), 5)
    job('report', 6, params=(('s1', 'tst_regression', 'desc'), ('c1', ('complete', 'partial'), 'desc')))
    job('promote', 7)
    return api


if __name__ == '__main__':
    create_jobs(api_select.ApiType.JENKINS)



# Copyright (c) 2019 Lars Hupfeldt Nielsen, Hupfeldt IT
# All rights reserved. This work is under a BSD license, see LICENSE.TXT.

import abc

from .propagation_types import  Propagation


class JobControlException(Exception):
    """Base for all Exceptions that may be raised by `jenkinsflow`."""
    def __init__(self, message, propagation=Propagation.NORMAL):
        super().__init__(message)
        self.propagation = propagation


class FlowTimeoutException(JobControlException):
    pass


class FlowScopeException(JobControlException):
    pass


class JobNotIdleException(JobControlException):
    pass


class MessageRedefinedException(JobControlException):
    pass


class JobControlFailException(JobControlException, metaclass=abc.ABCMeta):
    pass


class FailedSingleJobException(JobControlFailException):
    """Base for all Exceptions that may be raised by `jenkinsflow` because of failed job invocation or `build`."""
    def __init__(self, job, propagation):
        msg = f"Failed job: {repr(job)}, propagation: {propagation}."
        super().__init__(msg, propagation)


class MissingJobsException(JobControlFailException):
    def __init__(self, ex):
        super().__init__(str(ex), propagation=Propagation.NORMAL)


class FailedChildJobException(JobControlFailException):
    def __init__(self, flow_job, failed_child_job, propagation):
        msg = f"Failed child job in: {repr(flow_job)}, child job: {repr(failed_child_job)}, propagation: {propagation}."
        super().__init__(msg, propagation)


class FailedChildJobsException(JobControlFailException):
    def __init__(self, flow_job, failed_child_jobs, propagation):
        msg = f"Failed child jobs in: {repr(flow_job)}, child jobs: {repr(failed_child_jobs)}, propagation: {propagation}."
        super().__init__(msg, propagation)


class FinalResultException(JobControlFailException):
    def __init__(self, build_result):
        msg = f"Flow Unsuccessful: {build_result}"
        super().__init__(msg)
        self.result = build_result
